from setuptools import setup, find_packages

requirements = [
    'cobra>=0.5.4',
    'six',
    'pandas>=0.18.0',
    'fuzzywuzzy>=0.10.0',
    'requests',
    'tabulate',
    'numpy>=1.6'
]

try:
    with open('README.rst') as handle:
        description = handle.read()
except:
    description = ''

setup(
    name='cobrababel',
    version='0.1.3',
    packages=find_packages(),
    setup_requires=[],
    install_requires=requirements,
    tests_require=['pytest', 'lxml'],
    package_data={},
    author='Michael Mundy, Helena Mendes-Soares, Nicholas Chia',
    author_email='mundy.michael@mayo.edu',
    description='CobraBabel: COBRA Model Translator',
    long_description=description,
    license='BSD',
    keywords='metabolism biology optimization flux balance analysis fba',
    url='https://github.com/mmundy42/cobrababel',
    download_url='https://pypi.python.org/pypi/cobrababel',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
    platforms='GNU/Linux, Mac OS X >= 10.7, Microsoft Windows >= 7'
)
